/* j2kaudio.cpp

j2kaudio was first operational on July 01 2003

j2kaudio was written by Gregory Alan Hildstrom

j2kaudio License

Copyright (c) 2003 Gregory Alan Hildstrom

All rights reserved.

Permission is hereby granted, free of charge, to any person (the
"User") obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the
following conditions:

1.  The above copyright notices and this permission notice (which
includes the disclaimer below) shall be included in all copies or
substantial portions of the Software.

2.  The name of a copyright holder shall not be used to endorse or
promote products derived from the Software without specific prior
written permission.

THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS
LICENSE.  NO USE OF THE SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER
THIS DISCLAIMER.  THE SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
"AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.  IN NO
EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.  NO ASSURANCES ARE
PROVIDED BY THE COPYRIGHT HOLDERS THAT THE SOFTWARE DOES NOT INFRINGE
THE PATENT OR OTHER INTELLECTUAL PROPERTY RIGHTS OF ANY OTHER ENTITY.
EACH COPYRIGHT HOLDER DISCLAIMS ANY LIABILITY TO THE USER FOR CLAIMS
BROUGHT BY ANY OTHER ENTITY BASED ON INFRINGEMENT OF INTELLECTUAL
PROPERTY RIGHTS OR OTHERWISE.  AS A CONDITION TO EXERCISING THE RIGHTS
GRANTED HEREUNDER, EACH USER HEREBY ASSUMES SOLE RESPONSIBILITY TO SECURE
ANY OTHER INTELLECTUAL PROPERTY RIGHTS NEEDED, IF ANY.  THE SOFTWARE
IS NOT FAULT-TOLERANT AND IS NOT INTENDED FOR USE IN MISSION-CRITICAL
SYSTEMS, SUCH AS THOSE USED IN THE OPERATION OF NUCLEAR FACILITIES,
AIRCRAFT NAVIGATION OR COMMUNICATION SYSTEMS, AIR TRAFFIC CONTROL
SYSTEMS, DIRECT LIFE SUPPORT MACHINES, OR WEAPONS SYSTEMS, IN WHICH
THE FAILURE OF THE SOFTWARE OR SYSTEM COULD LEAD DIRECTLY TO DEATH,
PERSONAL INJURY, OR SEVERE PHYSICAL OR ENVIRONMENTAL DAMAGE ("HIGH
RISK ACTIVITIES").  THE COPYRIGHT HOLDERS SPECIFICALLY DISCLAIM ANY
EXPRESS OR IMPLIED WARRANTY OF FITNESS FOR HIGH RISK ACTIVITIES.

end j2kaudio License

I am not aware of any patented ideas present in this code. All algorithm
ideas for audio storage in the image file are my own. Please notify me immediately
if I am in violation of anything so I can remove it from the web.
Email with questions or concerns.
Yes I know that my code is not the cleanest.
Yes I know that there are different and better ways of doing things.
Please redirect flames to /dev/null.

The code to read and write wav files, which is included with
the j2kaudio source code, was written by
Dr Fred DePiero - CalPoly State University and is
open source, but not in the public domain.

j2kaudio uses the jasper implementation of the JPEG 2000 standard, which
is covered by its own open source public license.

*/

/*
j2kaudio is a program designed to leverage the wavelet compression of the JPEG 2000 standard
to compress audio data. The current design converts wav files to .jp2 JPEG2000 image files.
These image files are mostly compatible with .j2k files; I believe that the JPEG 2000 specifies
multiple different image and stream formats that may vary slightly by jpeg 2000 implementation.
j2kaudio currently uses the Jasper JPEG 2000 implementation and I am very impressed with their work.
Thank you to the Jasper developers.
Also thank you to Dr Fred DePiero; I am using his code to read and write wav files.
The original goal was to create a compression method for audio data
that can achieve high compression and/or high quality.

j2kaudio is fairly robust and has several
encoding methods to choose from. The first working encoding idea was tested with a 16-bit 44,100 Hz
2 channel wav file, which is standard CD quality audio; created with cdda2wav.
The first encoding method created a
2 component 16-bit grayscale image that was 44100 pixels wide x NumberOfSeconds pixels high.
This design is good because all of the information about the wav file is contained in the image. The
number of audio channels is the number of image components. The sample rate is the width of the image.
The number of seconds of audio is the height of the image. The number of samples is the height
multiplied by the width of the image.
Other methods that have been tried used a single image component and interleaved the audio channels
as alternating rows of pixels; another method that has been tried used separate image files to store
the individual audio channels, but both of these other methods require information not stored in the
image, which is not necessarily a problem.

One great advantage of using the JPEG 2000 standard is that it allows for lossy and losless
adaptive wavelet compression. This means that j2kaudio can perform lossy and lossless audio compression
with the same code and same algorithms. A good portion of the j2kaudio source code is matrix
manipulation to get the binary audio data into a highly compressible and managable two dimensional
or three dimensional matrix. I tried to use as much of the standard template library (STL) as possible.
Using STL helps reduce code size and eases dynamic memory management; it made my life easier.

The encodingrate input to j2kaudio is exactly the same as the encoding rate that must
be supplied to any other jpeg 2000 encoding engine. It spcifies the target reduction
in stored information. An encodingrate of 1 will perform lossless compression; it will
try to not lose any information. An encodingrate of 0.5 will try to lose 50% of the
information, but retain as much of the quality as possible with the current algorithms.
An encodingrate of 0.090726 performs almost the same size reduction as mp3 compression
using 128 as the bitrate.

Unfortunately j2kaudio does not currently sound as good as mp3 at moderate to heavy compression, but
it does do a more accurate job at recreating the original wav file. Unfortunately the recreation accuracy
does not matter if you can hear the difference with j2kaudio and not with mp3. j2kaudio can however
perform lossless compression, which mp3 cannot, which means that the recreated wav file is identical
to the original.

j2kaudio can also perform some extreme compression, like 100:1 or an encodingrate of 0.01. The
audio is recognizalbe, but the compression really suffers at the edge of the image. Sound recreated
from the edges of the image cannot be compressed as accurately as data in the middle of the image,
which leads to audible artifacts that occur every ImageWidth samples.

This program is a rediculous memory hog. It loads the entire wav file into memory at least twice
at the same time. It is a good idea to have several hundred megabytes of free ram to keep from swapping.
j2kaudio should run ok with small wav files and 512 MB of ram. I would recommend running it with 1 GB or more
for large wav files to keep from swapping.

There is currently no media player that will play these experimental .jp2 files. The process works like this:
j2kaudio 0 audio_in.wav image_out.jp2 0.1 - this will create an image file at roughly 10:1 compression.
j2kaudio 1 image_out.jp2 audio_recreated.wav - this will recreate a wav file from a properly create jp2 image file.
You can then play audio_in.wav and audio_recreated.wav to hear the original and the compressed versions.

During development I have been using jasper version jasper-1.700.2
Jasper is currently available from http://www.ece.uvic.ca/~mdadams/jasper/
You can also find the jasper page by performing a google search for "jasper jpeg 2000"
*/




#include <iostream>
#include <vector>
#include <string>
using namespace std;

#include "wav_in.h"
#include "wav_out.h"
#include <jasper/jasper.h>
#include "j2kaudio.h"










//functions that read in an n-channel by s-sample audio 2d matrix
vector< vector<short>*>* ReadWavToAudio2dVectorMatrix(char* filename, unsigned int & SampleRate){

	//attempt to load the wav file using the external wav code in wav_in.h
	WAV_IN infile(filename);
	SampleRate = (unsigned int)infile.get_sample_rate_hz();
	unsigned int BitsPerSample = infile.get_bits_per_sample();
	unsigned int AudioChannels = infile.get_num_channels();

	//check for CD quality stereo audio as a development starting point
	if(SampleRate != 44100 || BitsPerSample != 16 || (AudioChannels != 2 && AudioChannels != 2)){
		cerr << "Error: SampleRate: " << SampleRate << " BitsPerSample: " << BitsPerSample << " AudioChannels: " << AudioChannels << endl;
		cerr << "Error: The audio file is not standard CD quality audio, which is 44100 Hz, 16 bit, 2 channel audio." << endl;
		cerr.flush();
		return NULL;
	}//end if

	//define our matrix object, single audio channel, single sample, and current sample
	vector< vector<short>* > * Audio2dMatrix = NULL;
	vector<short>* AudioChannel = NULL;
	short datas16 = 0;
	unsigned int CurrentSample = 0;

	//allocate memory for our n-channel audio matrix
	Audio2dMatrix = new vector< vector<short>*>;

	//allocate memory for reading in our n audio channels
	for(unsigned int n=0; n < AudioChannels; n++){
		AudioChannel = new vector<short>;
		Audio2dMatrix->insert(Audio2dMatrix->end(), AudioChannel);
		//do not delete the memory pointed to by AudioChannel because we want to keep it
		AudioChannel = NULL;
	}//end for

	//read in some audio
	while(infile.more_data_available()) {

	 	//add the current sample to the correct audio channel object in the audio matrix
		for(unsigned int c=0; c < AudioChannels; c++){
			//I believe that the samples in a multi-channel wav file are interleaved
			//read in the current sample
			datas16 = infile.read_current_input_short();

			(*Audio2dMatrix)[c]->insert((*Audio2dMatrix)[c]->end(), datas16);
		}//end for
	}//end while



	return Audio2dMatrix;

}//end ReadWavToVectorMatrix







bool WriteAudio2dVectorMatrixToWav(char* filename, unsigned int SampleRate, vector< vector<short>*>* Audio2dMatrix){

	//attemp to load the wav file using the external wav code in wav_in.h
	unsigned int AudioChannels = Audio2dMatrix->size();
	WAV_OUT outfile(SampleRate, 16, AudioChannels);

	//check for CD quality stereo audio as a development starting point
	if(SampleRate != 44100 || 16 != 16 || (AudioChannels != 2 && AudioChannels != 2)){
		cerr << "Error: SampleRate: " << SampleRate << " BitsPerSample: " << 16 << " AudioChannels: " << AudioChannels << endl;
		cerr << "Error: The audio file is not standard CD quality audio, which is 44100 Hz, 16 bit, 2 channel audio." << endl;
		cerr.flush();
		return false;
	}//end if

	//define our matrix object, single audio channel, single sample, and current sample
	short datas16 = 0;
	unsigned int CurrentSample = 0;


	for(int sample = 0; sample < (*Audio2dMatrix)[0]->size(); sample++){
		for(int channel = 0; channel < AudioChannels; channel++){
			datas16 = (*(*Audio2dMatrix)[channel])[sample];
			outfile.write_current_output(datas16);
		}//end for
	}//end for


	outfile.save_wave_file(filename);

	return true;


}//end WriteAudio2dVectorMatrixToWav



















//functions that convert an n-channel by s-sample audio 2d matrix
//into a 3d matrix that is n-channels deep of sample rate by s-second 2d matrices
//these functions are used to get the linear audio channels into 2d matrices that
//can be processed by 2-dimensional image algorithms
//each individual channel will become a one second wide (sample rate) by s-second
//2d matrix
vector < vector< vector<short>*>*> * ConvertAudio2dVectorMatrixToAudio3dVectorMatrix(vector< vector<short>*>* Audio2dMatrix, unsigned int SampleRate){

	//setup initial variables
	vector < vector< vector<short>*>*> * Audio3dMatrix = new vector < vector< vector<short>*>*>;
	unsigned int AudioChannels = Audio2dMatrix->size();
	unsigned int SamplesPerChannel = (*Audio2dMatrix)[0]->size();
	unsigned int CurrentSample = 0;
	unsigned int Seconds = SamplesPerChannel / SampleRate;

	//create one 2d matrix for each audio channel
	for(int i=0; i<AudioChannels; i++)
		Audio3dMatrix->insert(Audio3dMatrix->end(), new vector< vector<short>*>);

	//populate the 2d matrices for each audio channel
	for(int channel=0; channel < AudioChannels; channel++){
		//start at the beginning of the audio stream
		CurrentSample = 0;
		//make the 2d matrix be one second wide, which happens to be the sample rate
		//scan across the audio for each second
		for(int sec=0; sec < Seconds; sec++){
			//insert a new array for each one second audio scan
			(*Audio3dMatrix)[channel]->insert((*Audio3dMatrix)[channel]->end(), new vector<short>);
			//populate the one second of data from the original 2d matrix
			for(int sample = 0; sample < SampleRate; sample++){
				//insert the audio sample
				(*(*Audio3dMatrix)[channel])[sec]->insert((*(*Audio3dMatrix)[channel])[sec]->end(), (*(*Audio2dMatrix)[channel])[CurrentSample]);
				//move on to the next original audio sample from the current channel
				CurrentSample++;
			}//end for
		}//end for
	}//end for

	return Audio3dMatrix;


}//end Convert2dAudioVectorMatrixTo3dAudioVectorMatrix






//this is the reverse function that takes a 3d audio matrix
//and transforms it back to the original n-channel x samples-per-channel 2d matrix
//functions that convert an n-channel by s-sample audio 2d matrix
//into a 3d matrix that is n-channels deep of sample rate by s-second 2d matrices
//these functions are used to get the linear audio channels into 2d matrices that
//can be processed by 2-dimensional image algorithms
//each individual channel will become a one second wide (sample rate) by s-second
//2d matrix
vector< vector<short>*>* ConvertAudio3dVectorMatrixToAudio2dVectorMatrix(vector < vector< vector<short>*>*> * Audio3dMatrix, unsigned int & SampleRate){

	//setup initial variables
	vector< vector<short>*>* Audio2dMatrix = new vector< vector<short>*>;
	unsigned int AudioChannels = Audio3dMatrix->size();
	SampleRate = (*(*Audio3dMatrix)[0])[0]->size();
	unsigned int Seconds = (*Audio3dMatrix)[0]->size();
	unsigned int SamplesPerChannel = SampleRate * Seconds;
	unsigned int CurrentSample = 0;


	//create one 1d array for each audio channel
	for(int i=0; i<AudioChannels; i++)
		Audio2dMatrix->insert(Audio2dMatrix->end(), new vector<short>);


	//populate the 2d matrices for each audio channel
	for(int channel=0; channel < AudioChannels; channel++){
		//start at the beginning of the audio stream
		CurrentSample = 0;
		//the 2d matrix is one second wide, which happens to be the sample rate
		//scan across the audio for each second

		for(int sec=0; sec < Seconds; sec++){

			//populate the new 2d matrix from the one second of data from the original 3d matrix
			for(int sample = 0; sample < SampleRate; sample++){
				//insert the audio sample
				(*Audio2dMatrix)[channel]->insert((*Audio2dMatrix)[channel]->end(), (*(*(*Audio3dMatrix)[channel])[sec])[sample]);
				//move on to the next original audio sample from the current channel
				CurrentSample++;
			}//end for

		}//end for

	}//end for

	return Audio2dMatrix;


}//end Convert2dAudioVectorMatrixTo3dAudioVectorMatrix
























//wav data to jp2 data function
//this function should receive a audio 3d vector matrix
//and output jp2 data to a named file
bool Audio3dVectorMatrixToJP2File(char* jp2filename, char* encodingrate, vector < vector< vector<short>*>*> * Audio3dVectorMatrix){

	jas_init();

	//the goal of this test code is to create a 2-component image from scratch and store it a s jp2 file
	//each component of the image will be 16 bit signed, just like a c++ short variable, which means
	//that this will be a 32-bit image.


	int numcmpts = Audio3dVectorMatrix->size();//number of image components, which should equal the number of audio channels
	jas_image_t * image = NULL;//the resultant image
	jas_image_cmptparm_t * compp;//two component parameters//only tested for 2 channel at the moment
	compp = new jas_image_cmptparm_t[numcmpts];
	jas_stream_t * outputstream = NULL;//jp2 stream
	vector<jas_matrix_t *> matrixdata;//component matrix data//only tested for 2 channel at the moment
	matrixdata.resize(numcmpts);
	unsigned int width = (*(*Audio3dVectorMatrix)[0])[0]->size();//width of the image, which should be one second of audio and equivalent to the sample rate
	unsigned int height = (*Audio3dVectorMatrix)[0]->size();//height of the image, which should be the number of seconds in the audio






	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general matrix to j2k steps
	//get data into 2 jas_matrix_t objects
	//get matrix objects into image as two components
	//encode the image into a stream object
	//write out the stream object to a j2k file







	//create components
	for(int i=0; i < numcmpts; i++){
		cout << "Creating and configuring new image component " << i << " parameters.\n";
		compp[i].tlx = 0;
		compp[i].tly = 0;
		compp[i].hstep = 1;
		compp[i].vstep = 1;
		compp[i].width = width;
		compp[i].height = height;
		compp[i].prec = 16;
		compp[i].sgnd = true;
		cout << "Just created and configured image component " << i << " parameters.\n";
	}//end for







	/* Create an image. */
	cout << "Creating image.\n";
	image = jas_image_create(numcmpts,compp, JAS_CLRSPC_SGRAY);
	if(image == NULL){
		cerr << "Error: image is NULL.\n";
		cerr.flush();
		cout << "Error: image is NULL.\n";
		cout.flush();
		return false;
	}//end if
	cout << "Just created a " << image->brx_ << " x " << image->bry_ << " image with " << image->numcmpts_ << " components.\n";





	cout << "Creating matrix data.\n";
	for(int i=0; i< numcmpts; i++){
		matrixdata[i] = jas_matrix_create(height, width);
		if(matrixdata[i] == NULL){
			cerr << "Error: matrixdata[" << i << "] is NULL.\n";
			cerr.flush();
			cout << "Error: matrixdata[" << i << "] is NULL.\n";
			cout.flush();
			return false;
		}//end if
	}//end for
	cout << "Just created matrix data.\n";






	cout << "Populating matrixdata.\n";
	for(int channel=0; channel < numcmpts; channel++){
		for(int sec = 0; sec < height; sec++){
			for(int sample = 0; sample < width; sample++){
				jas_matrix_set(matrixdata[channel], sec, sample, (*(*(*Audio3dVectorMatrix)[channel])[sec])[sample]);
				if(jas_matrix_get(matrixdata[channel], sec, sample) != (*(*(*Audio3dVectorMatrix)[channel])[sec])[sample]){
					cerr << "Error: matrix insertion did not work\n";
					return false;
				}//end if
			}//end for
		}//end for
	}//end for
	cout << "Just populated matrixdata.\n";





	//prelim cleanup to save some memory
	delete Audio3dVectorMatrix;
	Audio3dVectorMatrix = NULL;




	cout << "Writing matrixdata components to image.\n";
	/* Write a rectangular region of an image component. */
	for(int i=0; i < numcmpts; i++){
		if(jas_image_writecmpt(image,//image
			i,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			width,//jas_image_coord_t width
			height,//jas_image_coord_t height
  			matrixdata[i])){//jas_matrix_t *data
			cout << "Error: matrixdata not inserted into image object.\n";
			return false;
		}//end if
	}//end for
	cout << "Just wrote matrixdata components to image.\n";





	//prelim cleanup to save some memory
	for(int i=0; i < numcmpts; i++)
		jas_matrix_destroy(matrixdata[i]);







	/* Open a file as a stream. */
	if(!(outputstream = jas_stream_fopen(jp2filename, "rwb"))){
		cout << "Error: Could not open stream for writing.\n";
		return false;
	}//end if







	cout << "Encoding the image object to a stream object.\n";
	/* Write an image to a stream in a specified format. */
	cout << "Number of image components: " << jas_image_numcmpts(image) << endl;
	cout << "Maximum image components: " << image->maxcmpts_ << endl;
	if(image->cmpts_ == NULL)
		cout << "Per-component imformation is NULL.\n";
	if(image->cmprof_ == NULL)
		cout << "cmprof_ is NULL.\n";
	cout << "image->clrspc_: " << image->clrspc_ << endl;

	string encoderopts;
	//if the rate is valid then do lossy compression
	//else do lossless compression
	if(atof(encodingrate) > 0 && atof(encodingrate) < 1){
		encoderopts = "rate=";
		encoderopts.append(encodingrate);
		encoderopts.append("\n");
	}//end if
	encoderopts.append("mode=int\ncblkwidth=1024\ncblkheight=4 ");
	encoderopts[encoderopts.size()-1] = '\0';
	char* encopts = new char[encoderopts.size()];
	for(int i=0; i<encoderopts.size(); i++)
		encopts[i] = encoderopts[i];
	cout << "using encopts: " << encopts << endl;
	cout.flush();
	if(jas_image_encode(image, outputstream, jas_image_strtofmt("jp2"),encopts)){
		cout << "Error: could not encode image to stream.\n";
		return false;
	}//end if
	cout << "Just encoded the image object to a stream object.\n";


	//clean up
	jas_stream_close(outputstream);
	jas_image_destroy(image);


	return true;



}//end Audio3dVectorMatrixToJP2File























//wav data to jp2 data function
//this function should receive a audio 3d vector matrix
//and output jp2 data to a named file
vector < vector< vector<short>*>*> * JP2FileToAudio3dVectorMatrix(char* jp2filename){

	vector < vector< vector<short>*>*> * Audio3dVectorMatrix = NULL;


	jas_init();

	//the goal of this test code is to create a 2-component image from scratch and store it a s jp2 file
	//each component of the image will be 16 bit signed, just like a c++ short variable, which means
	//that this will be a 32-bit image.


	int numcmpts;//number of image components, which should equal the number of audio channels
	jas_image_t * image = NULL;//the resultant image
	jas_image_cmptparm_t * compp;//two component parameters//only tested for 2 channel at the moment
	compp = new jas_image_cmptparm_t[numcmpts];
	jas_stream_t * inputstream = NULL;//jp2 stream
	vector<jas_matrix_t *> matrixdata;//component matrix data//only tested for 2 channel at the moment
	unsigned int width;//width of the image, which should be one second of audio and equivalent to the sample rate
	unsigned int height;//height of the image, which should be the number of seconds in the audio






	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general j2k to matrix steps
	//read in the j2k file to a stream object
	//decode the stream object into an image object
	//check colorspace for GRAYscale
	//get image components into matrix objects
	//get 2 jas_matrix_t objects into Audio3dVectorMatrix





	/* Open a file as a stream. */
	if(!(inputstream = jas_stream_fopen(jp2filename, "rb"))){
		cout << "Error: Could not open stream for reading.\n";
		return NULL;
	}//end if






	cout << "Decoding the stream object to an image object.\n";
	/* read an image from a stream in a specified format. */
	if(!(image = jas_image_decode(inputstream, jas_image_strtofmt("jp2"),NULL))){
		cout << "Error: could not decode image from stream.\n";
		return NULL;
	}//end if
	cout << "Just decoded the image object from a stream object.\n";



	//clean up
	jas_stream_close(inputstream);





	//check the image to make sure it is in the grayscale colorspace
	if(jas_image_clrspc(image) != JAS_CLRSPC_SGRAY){
		cout << "Error: image is not grayscale.\n";
		return NULL;
	}//end if




	numcmpts = jas_image_numcmpts(image);
	matrixdata.resize(numcmpts);
	width = jas_image_width(image);
	height = jas_image_height(image);
	cout << "Decoding a " << numcmpts << " component image that is " << width << " x " << height << endl;









	cout << "Creating matrix data.\n";
	for(int i=0; i< numcmpts; i++){
		matrixdata[i] = jas_matrix_create(height, width);
		if(matrixdata[i] == NULL){
			cerr << "Error: matrixdata[" << i << "] is NULL.\n";
			cerr.flush();
			cout << "Error: matrixdata[" << i << "] is NULL.\n";
			cout.flush();
			return NULL;
		}//end if
	}//end for
	cout << "Just created matrix data.\n";









	cout << "Reading matrixdata components from image.\n";
	/* Write a rectangular region of an image component. */
	for(int i=0; i < numcmpts; i++){
		if(jas_image_readcmpt(image,//image
			i,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			width,//jas_image_coord_t width
			height,//jas_image_coord_t height
  			matrixdata[i])){//jas_matrix_t *data
			cout << "Error: matrixdata not inserted into image object.\n";
			return NULL;
		}//end if
	}//end for
	cout << "Just read matrixdata components from image.\n";







	//clean up
	jas_image_destroy(image);









	cout << "Populating Audio3dVectorMatrix.\n";
	Audio3dVectorMatrix = new vector < vector< vector<short>*>*>;
	for(int channel=0; channel < numcmpts; channel++){
		Audio3dVectorMatrix->insert(Audio3dVectorMatrix->end(), new vector< vector<short>*>);
		for(int sec = 0; sec < height; sec++){
			(*Audio3dVectorMatrix)[channel]->insert((*Audio3dVectorMatrix)[channel]->end(), new vector<short>);
			for(int sample = 0; sample < width; sample++){
				(*(*Audio3dVectorMatrix)[channel])[sec]->insert((*(*Audio3dVectorMatrix)[channel])[sec]->end(),jas_matrix_get(matrixdata[channel], sec, sample));
			}//end for
		}//end for
	}//end for
	cout << "Just populated Audio3dVectorMatrix.\n";












	//clean up
	for(int i=0; i < numcmpts; i++)
		jas_matrix_destroy(matrixdata[i]);



	return Audio3dVectorMatrix;



}//end JP2FileToAudio3dVectorMatrix



















//wav data to jp2 data function
//this function should receive a audio 3d vector matrix
//and output jp2 data to a named file
bool Audio3dVectorMatrixToJP2FileSmoothed(char* jp2filename, char* encodingrate, vector < vector< vector<short>*>*> * Audio3dVectorMatrix){

	jas_init();

	//the goal of this test code is to create a 2-component image from scratch and store it a s jp2 file
	//each component of the image will be 16 bit signed, just like a c++ short variable, which means
	//that this will be a 32-bit image.


	int numcmpts = Audio3dVectorMatrix->size();//number of image components, which should equal the number of audio channels
	jas_image_t * image = NULL;//the resultant image
	jas_image_cmptparm_t * compp;//two component parameters//only tested for 2 channel at the moment
	compp = new jas_image_cmptparm_t[numcmpts];
	jas_stream_t * outputstream = NULL;//jp2 stream
	vector<jas_matrix_t *> matrixdata;//component matrix data//only tested for 2 channel at the moment
	matrixdata.resize(numcmpts);
	unsigned int width = (*(*Audio3dVectorMatrix)[0])[0]->size();//width of the image, which should be one second of audio and equivalent to the sample rate
	unsigned int totalseconds = (*Audio3dVectorMatrix)[0]->size();//total number of audio seconds
	unsigned int height = (*Audio3dVectorMatrix)[0]->size()*2-1;//height of the image, which should be the number of seconds in the audio + smoothed pixels







	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general matrix to j2k steps
	//get data into 2 jas_matrix_t objects
	//get matrix objects into image as two components
	//encode the image into a stream object
	//write out the stream object to a j2k file







	//create components
	for(int i=0; i < numcmpts; i++){
		cout << "Creating and configuring new image component " << i << " parameters.\n";
		compp[i].tlx = 0;
		compp[i].tly = 0;
		compp[i].hstep = 1;
		compp[i].vstep = 1;
		compp[i].width = width;
		compp[i].height = height;
		compp[i].prec = 16;
		compp[i].sgnd = true;
		cout << "Just created and configured image component " << i << " parameters.\n";
	}//end for







	/* Create an image. */
	cout << "Creating image.\n";
	image = jas_image_create(numcmpts,compp, JAS_CLRSPC_SGRAY);
	if(image == NULL){
		cerr << "Error: image is NULL.\n";
		cerr.flush();
		cout << "Error: image is NULL.\n";
		cout.flush();
		return false;
	}//end if
	cout << "Just created a " << image->brx_ << " x " << image->bry_ << " image with " << image->numcmpts_ << " components.\n";





	cout << "Creating matrix data.\n";
	for(int i=0; i< numcmpts; i++){
		matrixdata[i] = jas_matrix_create(height, width);
		if(matrixdata[i] == NULL){
			cerr << "Error: matrixdata[" << i << "] is NULL.\n";
			cerr.flush();
			cout << "Error: matrixdata[" << i << "] is NULL.\n";
			cout.flush();
			return false;
		}//end if
	}//end for
	cout << "Just created matrix data.\n";






	cout << "Populating matrixdata.\n";
	for(int channel=0; channel < numcmpts; channel++){
		for(int sec = 0; sec < height; sec++){
			for(int sample = 0; sample < width; sample++){//insert the audio data
				if(sec % 2 == 0){//just insert the samples
					jas_matrix_set(matrixdata[channel], sec, sample, (*(*(*Audio3dVectorMatrix)[channel])[sec/2])[sample]);
					if(jas_matrix_get(matrixdata[channel], sec, sample) != (*(*(*Audio3dVectorMatrix)[channel])[sec/2])[sample]){
						cerr << "Error: matrix insertion did not work\n";
						return false;
					}//end if
				}//end if
				else{//insert the average samples to smooth out the image
					jas_matrix_set(matrixdata[channel], sec, sample, ((*(*(*Audio3dVectorMatrix)[channel])[sec/2])[sample]+(*(*(*Audio3dVectorMatrix)[channel])[sec/2+1])[sample])/2);
					if(jas_matrix_get(matrixdata[channel], sec, sample) != ((*(*(*Audio3dVectorMatrix)[channel])[sec/2])[sample]+(*(*(*Audio3dVectorMatrix)[channel])[sec/2+1])[sample])/2){
						cerr << "Error: matrix insertion did not work\n";
						return false;
					}//end if
				}//end else
			}//end for
		}//end for
	}//end for
	cout << "Just populated matrixdata.\n";





	//prelim cleanup to save some memory
	delete Audio3dVectorMatrix;
	Audio3dVectorMatrix = NULL;




	cout << "Writing matrixdata components to image.\n";
	/* Write a rectangular region of an image component. */
	for(int i=0; i < numcmpts; i++){
		if(jas_image_writecmpt(image,//image
			i,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			width,//jas_image_coord_t width
			height,//jas_image_coord_t height
  			matrixdata[i])){//jas_matrix_t *data
			cout << "Error: matrixdata not inserted into image object.\n";
			return false;
		}//end if
	}//end for
	cout << "Just wrote matrixdata components to image.\n";





	//prelim cleanup to save some memory
	for(int i=0; i < numcmpts; i++)
		jas_matrix_destroy(matrixdata[i]);







	/* Open a file as a stream. */
	if(!(outputstream = jas_stream_fopen(jp2filename, "rwb"))){
		cout << "Error: Could not open stream for writing.\n";
		return false;
	}//end if







	cout << "Encoding the image object to a stream object.\n";
	/* Write an image to a stream in a specified format. */
	cout << "Number of image components: " << jas_image_numcmpts(image) << endl;
	cout << "Maximum image components: " << image->maxcmpts_ << endl;
	if(image->cmpts_ == NULL)
		cout << "Per-component imformation is NULL.\n";
	if(image->cmprof_ == NULL)
		cout << "cmprof_ is NULL.\n";
	cout << "image->clrspc_: " << image->clrspc_ << endl;

	string encoderopts;
	//if the rate is valid then do lossy compression
	//else do lossless compression
	if(atof(encodingrate) > 0 && atof(encodingrate) < 1){
		encoderopts = "rate=";
		encoderopts.append(encodingrate);
		encoderopts.append("\n");
	}//end if
	encoderopts.append("mode=int\ncblkwidth=64\ncblkheight=64\nprcwidth=512\nprcheight=512 ");
	encoderopts[encoderopts.size()-1] = '\0';
	char* encopts = new char[encoderopts.size()];
	for(int i=0; i<encoderopts.size(); i++)
		encopts[i] = encoderopts[i];
	cout << "using encopts: " << encopts << endl;
	cout.flush();
	if(jas_image_encode(image, outputstream, jas_image_strtofmt("jp2"),encopts)){
		cout << "Error: could not encode image to stream.\n";
		return false;
	}//end if
	cout << "Just encoded the image object to a stream object.\n";


	//clean up
	jas_stream_close(outputstream);
	jas_image_destroy(image);


	return true;



}//end Audio3dVectorMatrixToJP2FileSmoothed























//wav data to jp2 data function
//this function should receive a audio 3d vector matrix
//and output jp2 data to a named file
vector < vector< vector<short>*>*> * JP2FileToAudio3dVectorMatrixSmoothed(char* jp2filename){

	vector < vector< vector<short>*>*> * Audio3dVectorMatrix = NULL;


	jas_init();

	//the goal of this test code is to create a 2-component image from scratch and store it a s jp2 file
	//each component of the image will be 16 bit signed, just like a c++ short variable, which means
	//that this will be a 32-bit image.


	int numcmpts;//number of image components, which should equal the number of audio channels
	jas_image_t * image = NULL;//the resultant image
	jas_image_cmptparm_t * compp;//two component parameters//only tested for 2 channel at the moment
	compp = new jas_image_cmptparm_t[numcmpts];
	jas_stream_t * inputstream = NULL;//jp2 stream
	vector<jas_matrix_t *> matrixdata;//component matrix data//only tested for 2 channel at the moment
	unsigned int width;//width of the image, which should be one second of audio and equivalent to the sample rate
	unsigned int totalseconds;//the total number of audio seconds
	unsigned int height;//height of the image, which should be the number of seconds in the audio






	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general j2k to matrix steps
	//read in the j2k file to a stream object
	//decode the stream object into an image object
	//check colorspace for GRAYscale
	//get image components into matrix objects
	//get 2 jas_matrix_t objects into Audio3dVectorMatrix





	/* Open a file as a stream. */
	if(!(inputstream = jas_stream_fopen(jp2filename, "rb"))){
		cout << "Error: Could not open stream for reading.\n";
		return NULL;
	}//end if






	cout << "Decoding the stream object to an image object.\n";
	/* read an image from a stream in a specified format. */
	if(!(image = jas_image_decode(inputstream, jas_image_strtofmt("jp2"),NULL))){
		cout << "Error: could not decode image from stream.\n";
		return NULL;
	}//end if
	cout << "Just decoded the image object from a stream object.\n";



	//clean up
	jas_stream_close(inputstream);





	//check the image to make sure it is in the grayscale colorspace
	if(jas_image_clrspc(image) != JAS_CLRSPC_SGRAY){
		cout << "Error: image is not grayscale.\n";
		return NULL;
	}//end if




	numcmpts = jas_image_numcmpts(image);
	matrixdata.resize(numcmpts);
	width = jas_image_width(image);
	height = jas_image_height(image);
	totalseconds = (height+1)/2;
	cout << "Decoding a " << numcmpts << " component image that is " << width << " x " << height << endl;









	cout << "Creating matrix data.\n";
	for(int i=0; i< numcmpts; i++){
		matrixdata[i] = jas_matrix_create(height, width);
		if(matrixdata[i] == NULL){
			cerr << "Error: matrixdata[" << i << "] is NULL.\n";
			cerr.flush();
			cout << "Error: matrixdata[" << i << "] is NULL.\n";
			cout.flush();
			return NULL;
		}//end if
	}//end for
	cout << "Just created matrix data.\n";









	cout << "Reading matrixdata components from image.\n";
	/* Write a rectangular region of an image component. */
	for(int i=0; i < numcmpts; i++){
		if(jas_image_readcmpt(image,//image
			i,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			width,//jas_image_coord_t width
			height,//jas_image_coord_t height
  			matrixdata[i])){//jas_matrix_t *data
			cout << "Error: matrixdata not extracted from image object.\n";
			return NULL;
		}//end if
	}//end for
	cout << "Just read matrixdata components from image.\n";







	//clean up
	jas_image_destroy(image);









	cout << "Populating Audio3dVectorMatrix.\n";
	Audio3dVectorMatrix = new vector < vector< vector<short>*>*>;
	for(int channel=0; channel < numcmpts; channel++){
		Audio3dVectorMatrix->insert(Audio3dVectorMatrix->end(), new vector< vector<short>*>);
		for(int sec = 0; sec < height; sec+=2){
			(*Audio3dVectorMatrix)[channel]->insert((*Audio3dVectorMatrix)[channel]->end(), new vector<short>);
			for(int sample = 0; sample < width; sample++){
				(*(*Audio3dVectorMatrix)[channel])[sec/2]->insert((*(*Audio3dVectorMatrix)[channel])[sec/2]->end(),jas_matrix_get(matrixdata[channel], sec, sample));
			}//end for
		}//end for
	}//end for
	cout << "Just populated Audio3dVectorMatrix.\n";












	//clean up
	for(int i=0; i < numcmpts; i++)
		jas_matrix_destroy(matrixdata[i]);



	return Audio3dVectorMatrix;



}//end JP2FileToAudio3dVectorMatrixSmoothed




























//wav data to jp2 data function
//this function should receive a audio 3d vector matrix
//and output jp2 data to a named file
bool Audio3dVectorMatrixToJP2FileInterleaved(char* jp2filename, char* encodingrate, vector< vector< vector<short>*>*> * Audio3dVectorMatrix){

	jas_init();

	//the goal of this test code is to create a 2-component image from scratch and store it a s jp2 file
	//each component of the image will be 16 bit signed, just like a c++ short variable, which means
	//that this will be a 32-bit image.


	unsigned int channels = Audio3dVectorMatrix->size();
	int numcmpts = 1;//number of image components, which should be one because all audio channels are stored in the first component
	jas_image_t * image = NULL;//the resultant image
	jas_image_cmptparm_t * compp;//two component parameters//only tested for 2 channel at the moment
	compp = new jas_image_cmptparm_t[numcmpts];
	jas_stream_t * outputstream = NULL;//jp2 stream
	vector<jas_matrix_t *> matrixdata;//component matrix data//only tested for 2 channel at the moment
	matrixdata.resize(numcmpts);
	unsigned int width = (*(*Audio3dVectorMatrix)[0])[0]->size();//width of the image, which should be one second of audio and equivalent to the sample rate
	unsigned int height = (*Audio3dVectorMatrix)[0]->size() * channels;//height of the image, which should be the number of seconds in the audio * number of channels






	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general matrix to j2k steps
	//get data into 2 jas_matrix_t objects
	//get matrix objects into image as two components
	//encode the image into a stream object
	//write out the stream object to a j2k file







	//create components
	for(int i=0; i < numcmpts; i++){
		cout << "Creating and configuring new image component " << i << " parameters.\n";
		compp[i].tlx = 0;
		compp[i].tly = 0;
		compp[i].hstep = 1;
		compp[i].vstep = 1;
		compp[i].width = width;
		compp[i].height = height;
		compp[i].prec = 16;
		compp[i].sgnd = true;
		cout << "Just created and configured image component " << i << " parameters.\n";
	}//end for







	/* Create an image. */
	cout << "Creating image.\n";
	image = jas_image_create(numcmpts,compp, JAS_CLRSPC_SGRAY);
	if(image == NULL){
		cerr << "Error: image is NULL.\n";
		cerr.flush();
		cout << "Error: image is NULL.\n";
		cout.flush();
		return false;
	}//end if
	cout << "Just created a " << image->brx_ << " x " << image->bry_ << " image with " << image->numcmpts_ << " components.\n";





	cout << "Creating matrix data.\n";
	for(int i=0; i< numcmpts; i++){
		matrixdata[i] = jas_matrix_create(height, width);
		if(matrixdata[i] == NULL){
			cerr << "Error: matrixdata[" << i << "] is NULL.\n";
			cerr.flush();
			cout << "Error: matrixdata[" << i << "] is NULL.\n";
			cout.flush();
			return false;
		}//end if
	}//end for
	cout << "Just created matrix data.\n";






	cout << "Populating matrixdata.\n";
	for(int sec = 0; sec < height/channels; sec++){
		for(int channel=0; channel < channels; channel++){
			for(int sample = 0; sample < width; sample++){
				jas_matrix_set(matrixdata[0], sec*channels + channel, sample, (*(*(*Audio3dVectorMatrix)[channel])[sec])[sample]);
				if(jas_matrix_get(matrixdata[0], sec*channels + channel, sample) != (*(*(*Audio3dVectorMatrix)[channel])[sec])[sample]){
					cerr << "Error: matrix insertion did not work\n";
					return false;
				}//end if
			}//end for
		}//end for
	}//end for
	cout << "Just populated matrixdata.\n";





	//prelim cleanup to save some memory
	delete Audio3dVectorMatrix;
	Audio3dVectorMatrix = NULL;




	cout << "Writing matrixdata components to image.\n";
	/* Write a rectangular region of an image component. */
	for(int i=0; i < numcmpts; i++){
		if(jas_image_writecmpt(image,//image
			i,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			width,//jas_image_coord_t width
			height,//jas_image_coord_t height
  			matrixdata[i])){//jas_matrix_t *data
			cout << "Error: matrixdata not inserted into image object.\n";
			return false;
		}//end if
	}//end for
	cout << "Just wrote matrixdata components to image.\n";





	//prelim cleanup to save some memory
	for(int i=0; i < numcmpts; i++)
		jas_matrix_destroy(matrixdata[i]);







	/* Open a file as a stream. */
	if(!(outputstream = jas_stream_fopen(jp2filename, "rwb"))){
		cout << "Error: Could not open stream for writing.\n";
		return false;
	}//end if







	cout << "Encoding the image object to a stream object.\n";
	/* Write an image to a stream in a specified format. */
	cout << "Number of image components: " << jas_image_numcmpts(image) << endl;
	cout << "Maximum image components: " << image->maxcmpts_ << endl;
	if(image->cmpts_ == NULL)
		cout << "Per-component imformation is NULL.\n";
	if(image->cmprof_ == NULL)
		cout << "cmprof_ is NULL.\n";
	cout << "image->clrspc_: " << image->clrspc_ << endl;

	string encoderopts;
	//if the rate is valid then do lossy compression
	//else do lossless compression
	if(atof(encodingrate) > 0 && atof(encodingrate) < 1){
		encoderopts = "rate=";
		encoderopts.append(encodingrate);
		encoderopts.append("\n");
	}//end if
	encoderopts.append("mode=int\ncblkwidth=1024\ncblkheight=4 ");
	encoderopts[encoderopts.size()-1] = '\0';
	char* encopts = new char[encoderopts.size()];
	for(int i=0; i<encoderopts.size(); i++)
		encopts[i] = encoderopts[i];
	cout << "using encopts: " << encopts << endl;
	cout.flush();
	if(jas_image_encode(image, outputstream, jas_image_strtofmt("jp2"),encopts)){
		cout << "Error: could not encode image to stream.\n";
		return false;
	}//end if
	cout << "Just encoded the image object to a stream object.\n";


	//clean up
	jas_stream_close(outputstream);
	jas_image_destroy(image);


	return true;



}//end Audio3dVectorMatrixToJP2FileInterleaved























//wav data to jp2 data function
//this function should receive a audio 3d vector matrix
//and output jp2 data to a named file
vector< vector< vector<short>*>*> * JP2FileToAudio3dVectorMatrixInterleaved(char* jp2filename, unsigned int channels){

	vector< vector< vector<short>*>*> * Audio3dVectorMatrix = NULL;


	jas_init();

	//the goal of this test code is to create a 2-component image from scratch and store it a s jp2 file
	//each component of the image will be 16 bit signed, just like a c++ short variable, which means
	//that this will be a 32-bit image.


	int numcmpts = 1;//number of image components, which should equal the number of audio channels
	jas_image_t * image = NULL;//the resultant image
	jas_image_cmptparm_t * compp;//two component parameters//only tested for 2 channel at the moment
	compp = new jas_image_cmptparm_t[numcmpts];
	jas_stream_t * inputstream = NULL;//jp2 stream
	vector<jas_matrix_t *> matrixdata;//component matrix data//only tested for 2 channel at the moment
	matrixdata.resize(numcmpts);
	unsigned int width;//width of the image, which should be one second of audio and equivalent to the sample rate
	unsigned int height;//height of the image, which should be the number of seconds in the audio






	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general j2k to matrix steps
	//read in the j2k file to a stream object
	//decode the stream object into an image object
	//check colorspace for GRAYscale
	//get image components into matrix objects
	//get 2 jas_matrix_t objects into Audio3dVectorMatrix





	/* Open a file as a stream. */
	if(!(inputstream = jas_stream_fopen(jp2filename, "rb"))){
		cout << "Error: Could not open stream for reading.\n";
		return NULL;
	}//end if






	cout << "Decoding the stream object to an image object.\n";
	/* read an image from a stream in a specified format. */
	if(!(image = jas_image_decode(inputstream, jas_image_strtofmt("jp2"),NULL))){
		cout << "Error: could not decode image from stream.\n";
		return NULL;
	}//end if
	cout << "Just decoded the image object from a stream object.\n";



	//clean up
	jas_stream_close(inputstream);





	//check the image to make sure it is in the grayscale colorspace
	if(jas_image_clrspc(image) != JAS_CLRSPC_SGRAY){
		cout << "Error: image is not grayscale.\n";
		return NULL;
	}//end if




	numcmpts = jas_image_numcmpts(image);
	width = jas_image_width(image);
	height = jas_image_height(image);
	cout << "Decoding a " << numcmpts << " component image that is " << width << " x " << height << endl;









	cout << "Creating matrix data.\n";
	for(int i=0; i< numcmpts; i++){
		matrixdata[i] = jas_matrix_create(height, width);
		if(matrixdata[i] == NULL){
			cerr << "Error: matrixdata[" << i << "] is NULL.\n";
			cerr.flush();
			cout << "Error: matrixdata[" << i << "] is NULL.\n";
			cout.flush();
			return NULL;
		}//end if
	}//end for
	cout << "Just created matrix data.\n";









	cout << "Reading matrixdata components from image.\n";
	/* Write a rectangular region of an image component. */
	for(int i=0; i < numcmpts; i++){
		if(jas_image_readcmpt(image,//image
			i,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			width,//jas_image_coord_t width
			height,//jas_image_coord_t height
  			matrixdata[i])){//jas_matrix_t *data
			cout << "Error: matrixdata not inserted into image object.\n";
			return NULL;
		}//end if
	}//end for
	cout << "Just read matrixdata components from image.\n";







	//clean up
	jas_image_destroy(image);









	cout << "Populating Audio3dVectorMatrix.\n";
	Audio3dVectorMatrix = new vector < vector< vector<short>*>*>;
	for(int channel=0; channel < channels; channel++){
		Audio3dVectorMatrix->insert(Audio3dVectorMatrix->end(), new vector< vector<short>*>);
		for(int sec = 0; sec < height/channels; sec++){
			(*Audio3dVectorMatrix)[channel]->insert((*Audio3dVectorMatrix)[channel]->end(), new vector<short>);
			for(int sample = 0; sample < width; sample++){
				(*(*Audio3dVectorMatrix)[channel])[sec]->insert((*(*Audio3dVectorMatrix)[channel])[sec]->end(),jas_matrix_get(matrixdata[0], sec*channels + channel, sample));
			}//end for
		}//end for
	}//end for
	cout << "Just populated Audio3dVectorMatrix.\n";












	//clean up
	for(int i=0; i < numcmpts; i++)
		jas_matrix_destroy(matrixdata[i]);



	return Audio3dVectorMatrix;



}//end JP2FileToAudio3dVectorMatrixInterleaved



















//wav data to jp2 data function
//this function should receive a audio 3d vector matrix
//and output jp2 data to a named file
bool Audio3dVectorMatrixToJP2FileSImage(char* jp2filenameprefix, char* encodingrate, vector < vector< vector<short>*>*> * Audio3dVectorMatrix){

	jas_init();

	//the goal of this test code is to create a 1-component image from scratch and store it as many jp2 files
	//each component of the image will be 16 bit signed, just like a c++ short variable, which means
	//that these images will be a 16-bit grayscale images.


	int numcmpts = 1;//number of image components, which should be one
	jas_image_t * image = NULL;//the resultant image
	jas_image_cmptparm_t * compp;//two component parameters//only tested for 2 channel at the moment
	compp = new jas_image_cmptparm_t[numcmpts];
	jas_stream_t * outputstream = NULL;//jp2 stream
	vector<jas_matrix_t *> matrixdata;//component matrix data//only tested for 2 channel at the moment
	matrixdata.resize(numcmpts);
	unsigned int width = (*(*Audio3dVectorMatrix)[0])[0]->size();//width of the image, which should be one second of audio and equivalent to the sample rate
	unsigned int height = Audio3dVectorMatrix->size();//height of the image, which should be the number of channels in the audio
	unsigned int currentsecond = 0;






	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general matrix to j2k steps
	//get data into 2 jas_matrix_t objects
	//get matrix objects into image as two components
	//encode the image into a stream object
	//write out the stream object to a j2k file







	//create components
	for(int i=0; i < numcmpts; i++){
		cout << "Creating and configuring new image component " << i << " parameters.\n";
		compp[i].tlx = 0;
		compp[i].tly = 0;
		compp[i].hstep = 1;
		compp[i].vstep = 1;
		compp[i].width = width;
		compp[i].height = height;
		compp[i].prec = 16;
		compp[i].sgnd = true;
		cout << "Just created and configured image component " << i << " parameters.\n";
	}//end for









	while(currentsecond < (*Audio3dVectorMatrix)[0]->size()){//while there are more seconds of audio left

	cout << "Entered beginning of while loop\n";
	cout.flush();








	/* Create an image. */
	cout << "Creating image.\n";
	cout.flush();
	image = jas_image_create(numcmpts,compp, JAS_CLRSPC_SGRAY);
	if(image == NULL){
		cerr << "Error: image is NULL.\n";
		cerr.flush();
		cout << "Error: image is NULL.\n";
		cout.flush();
		return false;
	}//end if
	cout << "Just created a " << image->brx_ << " x " << image->bry_ << " image with " << image->numcmpts_ << " components.\n";





	cout << "Creating matrix data.\n";
	for(int i=0; i< numcmpts; i++){
		matrixdata[i] = jas_matrix_create(height, width);
		if(matrixdata[i] == NULL){
			cerr << "Error: matrixdata[" << i << "] is NULL.\n";
			cerr.flush();
			cout << "Error: matrixdata[" << i << "] is NULL.\n";
			cout.flush();
			return false;
		}//end if
	}//end for
	cout << "Just created matrix data.\n";






	cout << "Populating matrixdata.\n";
		for(int channel = 0; channel < height; channel++){
			for(int sample = 0; sample < width; sample++){
				jas_matrix_set(matrixdata[0], channel, sample, (*(*(*Audio3dVectorMatrix)[channel])[currentsecond])[sample]);
				if(jas_matrix_get(matrixdata[0], channel, sample) != (*(*(*Audio3dVectorMatrix)[channel])[currentsecond])[sample]){
					cerr << "Error: matrix insertion did not work\n";
					return false;
				}//end if
			}//end for
		}//end for
	cout << "Just populated matrixdata.\n";










	cout << "Writing matrixdata components to image.\n";
	/* Write a rectangular region of an image component. */
	for(int i=0; i < numcmpts; i++){
		if(jas_image_writecmpt(image,//image
			i,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			width,//jas_image_coord_t width
			height,//jas_image_coord_t height
  			matrixdata[i])){//jas_matrix_t *data
			cout << "Error: matrixdata not inserted into image object.\n";
			return false;
		}//end if
	}//end for
	cout << "Just wrote matrixdata components to image.\n";





	//prelim cleanup to save some memory
	for(int i=0; i < numcmpts; i++)
		jas_matrix_destroy(matrixdata[i]);








	//build sequential filenames
	char* jp2filename = NULL;
	char numberstring[40];
	sprintf(numberstring,"%d",currentsecond);
	string jp2filenamestring = jp2filenameprefix;
	jp2filenamestring.append(numberstring);
	jp2filenamestring.append(".jp2 ");
	jp2filenamestring[jp2filenamestring.size()-1] = '\0';
	jp2filename = new char[jp2filenamestring.size()];
	strcpy(jp2filename, jp2filenamestring.data());












	/* Open a file as a stream. */
	if(!(outputstream = jas_stream_fopen(jp2filename, "rwb"))){
		cout << "Error: Could not open stream for writing.\n";
		return false;
	}//end if







	cout << "Encoding the image object to a stream object.\n";
	/* Write an image to a stream in a specified format. */
	cout << "Number of image components: " << jas_image_numcmpts(image) << endl;
	cout << "Maximum image components: " << image->maxcmpts_ << endl;
	if(image->cmpts_ == NULL)
		cout << "Per-component imformation is NULL.\n";
	if(image->cmprof_ == NULL)
		cout << "cmprof_ is NULL.\n";
	cout << "image->clrspc_: " << image->clrspc_ << endl;

	string encoderopts;
	//if the rate is valid then do lossy compression
	//else do lossless compression
	if(atof(encodingrate) > 0 && atof(encodingrate) < 1){
		encoderopts = "rate=";
		encoderopts.append(encodingrate);
		encoderopts.append("\n");
	}//end if
	encoderopts.append("mode=int ");
	encoderopts[encoderopts.size()-1] = '\0';
	char* encopts = new char[encoderopts.size()];
	for(int i=0; i<encoderopts.size(); i++)
		encopts[i] = encoderopts[i];
	cout << "using encopts: " << encopts << endl;
	cout.flush();
	if(jas_image_encode(image, outputstream, jas_image_strtofmt("jp2"),encopts)){
		cout << "Error: could not encode image to stream.\n";
		return false;
	}//end if
	cout << "Just encoded the image object to a stream object.\n";

	cout << "About to clean up.\n";
	cout.flush();
	//clean up
	jas_stream_close(outputstream);
	jas_image_destroy(image);
	image = NULL;
	cout << "Just cleaned up.\n";
	cout.flush();



	currentsecond++;
	cout << "Just incremented current second to " << currentsecond << endl;
	cout.flush();

	}//end while more seconds left


	cout << "Finished while loop\n";
	cout.flush();

	//prelim cleanup to save some memory
	delete Audio3dVectorMatrix;
	Audio3dVectorMatrix = NULL;






	return true;



}//end Audio3dVectorMatrixToJP2FileSImage























//wav data to jp2 data function
//this function should receive a audio 3d vector matrix
//and output jp2 data to a named file
vector < vector< vector<short>*>*> * JP2FileToAudio3dVectorMatrixSImage(char* jp2filenameprefix){

	vector < vector< vector<short>*>*> * Audio3dVectorMatrix = NULL;


	jas_init();

	//the goal of this test code is to create a 2-component image from scratch and store it a s jp2 file
	//each component of the image will be 16 bit signed, just like a c++ short variable, which means
	//that this will be a 32-bit image.


	int numcmpts;//number of image components, which should equal the number of audio channels
	jas_image_t * image = NULL;//the resultant image
	jas_image_cmptparm_t * compp;//two component parameters//only tested for 2 channel at the moment
	compp = new jas_image_cmptparm_t[numcmpts];
	jas_stream_t * inputstream = NULL;//jp2 stream
	vector<jas_matrix_t *> matrixdata;//component matrix data//only tested for 2 channel at the moment
	unsigned int width;//width of the image, which should be one second of audio and equivalent to the sample rate
	unsigned int height;//height of the image, which should be the number of seconds in the audio






	////////////////////////
	//This is my best preliminary guess as to the steps involved.
	//general j2k to matrix steps
	//read in the j2k file to a stream object
	//decode the stream object into an image object
	//check colorspace for GRAYscale
	//get image components into matrix objects
	//get 2 jas_matrix_t objects into Audio3dVectorMatrix











	unsigned int currentsecond = 0;











	//build sequential filenames
	char* jp2filename = NULL;
	char numberstring[40];
	sprintf(numberstring,"%d",currentsecond);
	string jp2filenamestring = jp2filenameprefix;
	jp2filenamestring.append(numberstring);
	jp2filenamestring.append(".jp2 ");
	jp2filenamestring[jp2filenamestring.size()-1] = '\0';
	jp2filename = new char[jp2filenamestring.size()];
	strcpy(jp2filename, jp2filenamestring.data());


	//must create the 3d matrix first
	Audio3dVectorMatrix = new vector < vector< vector<short>*>*>;



	/* Open a file as a stream. */
	while(inputstream = jas_stream_fopen(jp2filename, "rb")){






	cout << "Decoding the stream object to an image object.\n";
	/* read an image from a stream in a specified format. */
	if(!(image = jas_image_decode(inputstream, jas_image_strtofmt("jp2"),NULL))){
		cout << "Error: could not decode image from stream.\n";
		return NULL;
	}//end if
	cout << "Just decoded the image object from a stream object.\n";



	//clean up
	jas_stream_close(inputstream);





	//check the image to make sure it is in the grayscale colorspace
	if(jas_image_clrspc(image) != JAS_CLRSPC_SGRAY){
		cout << "Error: image is not grayscale.\n";
		return NULL;
	}//end if




	numcmpts = jas_image_numcmpts(image);
	matrixdata.resize(numcmpts);
	width = jas_image_width(image);
	height = jas_image_height(image);
	cout << "Decoding a " << numcmpts << " component image that is " << width << " x " << height << endl;









	cout << "Creating matrix data.\n";
	for(int i=0; i< numcmpts; i++){
		matrixdata[i] = jas_matrix_create(height, width);
		if(matrixdata[i] == NULL){
			cerr << "Error: matrixdata[" << i << "] is NULL.\n";
			cerr.flush();
			cout << "Error: matrixdata[" << i << "] is NULL.\n";
			cout.flush();
			return NULL;
		}//end if
	}//end for
	cout << "Just created matrix data.\n";









	cout << "Reading matrixdata components from image.\n";
	/* Write a rectangular region of an image component. */
	for(int i=0; i < numcmpts; i++){
		if(jas_image_readcmpt(image,//image
			i,//component number
  			0,//jas_image_coord_t x
			0,//jas_image_coord_t y
			width,//jas_image_coord_t width
			height,//jas_image_coord_t height
  			matrixdata[i])){//jas_matrix_t *data
			cout << "Error: matrixdata not inserted into image object.\n";
			return NULL;
		}//end if
	}//end for
	cout << "Just read matrixdata components from image.\n";







	//clean up
	jas_image_destroy(image);









	cout << "Populating Audio3dVectorMatrix.\n";
	for(int channel=0; channel < height; channel++){
		if(currentsecond == 0){
			Audio3dVectorMatrix->insert(Audio3dVectorMatrix->end(), new vector< vector<short>*>);
		}//end if
			(*Audio3dVectorMatrix)[channel]->insert((*Audio3dVectorMatrix)[channel]->end(), new vector<short>);
			for(int sample = 0; sample < width; sample++){
				(*(*Audio3dVectorMatrix)[channel])[currentsecond]->insert((*(*Audio3dVectorMatrix)[channel])[currentsecond]->end(),jas_matrix_get(matrixdata[0], channel, sample));
			}//end for
	}//end for
	cout << "Just populated Audio3dVectorMatrix.\n";




	//clean up
	for(int i=0; i < numcmpts; i++)
		jas_matrix_destroy(matrixdata[i]);




	currentsecond++;

	//build sequential filenames
	sprintf(numberstring,"%d",currentsecond);
	jp2filenamestring = jp2filenameprefix;
	jp2filenamestring.append(numberstring);
	jp2filenamestring.append(".jp2 ");
	jp2filenamestring[jp2filenamestring.size()-1] = '\0';
	jp2filename = new char[jp2filenamestring.size()];
	strcpy(jp2filename, jp2filenamestring.data());



	}//end while




	return Audio3dVectorMatrix;



}//end JP2FileToAudio3dVectorMatrixSImage























